from flask import current_app
from importlib import import_module
from polzybackend.mediators import Policy, Antrag
#from polzybackend.models import Activity


def import_class(name):
    #
    # imports class from string 'name' 
    #

    components = name.split('.')
    module_name = '.'.join(components[:-1])
    mod = import_module(module_name)
    cls = getattr(mod, components[-1])
    return cls


'''
def activity_class(activity_type_class):
    #
    # returns activity class
    #

    if current_app.config.get(activity_type_class):
        return import_class(current_app.config.get(activity_type_class))

    return Activity
'''

def all_stages():
    #
    # returns list of all stages
    #

    if current_app.config.get('CLASSNAME_STAGES'):
        lClass = import_class(current_app.config.get('CLASSNAME_STAGES'))()
        return lClass.getAllStages

    raise Exception('Method to Get Stages NOT defined')


def record_login(user):
    if current_app.config.get('LOGIN_RECORDER'):
        import_class(current_app.config.get('LOGIN_RECORDER'))(user).record_login_activity()


def permissions(user):
    #
    # returns user's permissions
    #
    record_login(user)
    if current_app.config.get('METHOD_GET_PERMISSIONS'):
        permissions_method = import_class(current_app.config['METHOD_GET_PERMISSIONS'])
        return permissions_method(user)

    # default method
    return user.get_permissions()


def policy_class():
    #
    # returns policy class of the current implementation
    #

    if current_app.config.get('CLASSNAME_POLICY'):
        return import_class(current_app.config.get('CLASSNAME_POLICY'))

    return Policy


def antrag_class():
    #
    # returns antrag class of the current implementation
    #

    if current_app.config.get('CLASSNAME_ANTRAG'):
        return import_class(current_app.config['CLASSNAME_ANTRAG'])

    return Antrag

def antrag_products():
    #
    # returns list of avalaibale products
    #

    if current_app.config.get('CLASSNAME_PRODUCTS'):
        return import_class(current_app.config['CLASSNAME_PRODUCTS'])

    raise Exception('Class NOT difined to derive products')

def gamification_ranking(user):
    #
    # returns user's gamification ranking
    #

    if current_app.config.get('CLASSNAME_GAMIFICATION_HITLIST'):
        lClass = import_class(current_app.config.get('CLASSNAME_GAMIFICATION_HITLIST'))()
        return lClass.deriveUserRanking(user)

    raise Exception("Class NOT difined to derive user's ranking")



import requests
import json
try:
    from polzybackend import messenger
    noFlaskMode = False
except ImportError:
    # We're not in Flask
    noFlaskMode = True

base_url = "http://0.0.0.0:9000/"
token = None
header = None


def get_token():
    # fixme: This is the old gamification implementation, which should be completely removed
    return None

    global token, header
    res = requests.post(base_url + "auth/login", json={"email": "admin@fasifu.com", "password": "admin@fasifu"})
    token = json.loads(res.content).get("token")
    header = {'X-Auth-Token': token}
    if not token:
        print("Can't get AuthUser token!!")


def update_achievement(variable="policy_inquiry", user_id="1"):
    # fixme: This is the old gamification implementation, which should be completely removed
    return None

    if not token:
        try:
            get_token()
        except Exception as e:
            print("can't get user token. in GamificationHandler.py")

    if token:
        try:
            res = requests.post(base_url + f"increase_value/{variable}/{str(user_id)}", headers=header,
                                data={"value": "1"})
            json_response = json.loads(res.content)
            if json_response["achievements"]:
                announceMessage(
                    f"Congratulations! you have achieved {json_response['achievements'][0]['internal_name']}",
                    duration=5000
                )
        except Exception as ex:
            print(ex)


def announceMessage(message, duration=3000, level='default', horizontal='left', vertical='bottom'):
    if noFlaskMode:
        return
    msg = json.dumps({
        'text': message,
        'autoHideDuration': duration,
        'variant': level,
        'anchorOrigin': {
            'horizontal': horizontal,
            'vertical': vertical,
        }
    })
    messenger.announce(f"data: {msg}\n\n")




from flask import jsonify, request, current_app
from polzybackend.gamification import bp
from polzybackend.models import GamificationBadge, GamificationBadgeType
from polzybackend import auth
from polzybackend.utils.import_utils import gamification_ranking


@bp.route('/badges')
@auth.login_required
def badges():
    #
    # returns list of all user badges
    #

    try:
        badges = [badge.to_json() for badge in auth.current_user().badges]
        return jsonify(badges), 200
    except Exception as e:
        current_app.logger.exception(f'Faild to get Gamification Badges: {e}')
        return jsonify({'error': 'Bad Request'}), 400


def filter_badge(x, badge):
    if x.type.name == badge.get('type') and x.level.name == badge.get('level'):
        return x


@bp.route('/badges/seen', methods=['POST'])
@auth.login_required
def make_badge_seen():
    #
    # makes the bage as seen
    #

    try:

        # get request data and badge
        data = request.get_json()
        badge = data.get('badge')
        if badge is None:
            raise Exception('Badge data not found in request')

        ###### DEUG OUTPUT
        print("\n**** Badge Seen:")
        import json
        print(json.dumps(data, indent=2))

        # find badge in user badges
        userBadge = list(filter(lambda x: filter_badge(x, badge), auth.current_user().badges))
        if len(userBadge) == 0:
            raise Exception(f"Badge Type {badge.get('type')} not found in user's badges")

        # update badge instance
        userBadge[0].set_seen()

        # return updated lit of badges
        return jsonify([badge.to_json() for badge in auth.current_user().badges]), 200

    except Exception as e:
        current_app.logger.exception(f'Making Gamification Badge seen fails: Badge={badge}\n{e}')
        return jsonify({'error': 'Bad Request'}), 400


@bp.route('/badges/types')
@auth.login_required
def badge_types():
    #
    # returns list of available types of badges
    #

    try:
        badge_types = GamificationBadgeType.query.order_by('id').all()
        return jsonify([badge.to_json() for badge in badge_types]), 200
    except Exception as e:
        current_app.logger.exception(f'Faild to get Gamification Badge Types: {e}')
    
    return jsonify({'error': 'Bad Request'}), 400


@bp.route('/rankings')
@auth.login_required
def rankings():
    #
    # returns user's rankings
    #

    try:
        return jsonify(gamification_ranking(auth.current_user())), 200
    except Exception as e:
        current_app.logger.exception(f"Faild to get user's gamification rankings: {e}")
    
    return jsonify({'error': 'Bad Request'}), 400





from flask import Response
from polzybackend.announce import bp
from polzybackend import auth, db
from polzybackend import messenger
from polzybackend.models import ToastNotifications, GamificationBadge
import json
from logging import getLogger

logger = getLogger("PyC")

@bp.route('/listen', methods=['GET'])
def listen():

    def stream():
        messages = messenger.listen()  # returns a queue.Queue
        while True:
            msg = messages.get()  # blocks until a new message arrives
            yield msg

    return Response(
        stream(),
        headers={
            "Access-Control-Allow-Origin": "*",
            "Access-Control-Allow-Headers": "Origin, X-Requested-With, Content-Type, Accept",
            "X-Accel-Buffering": "no",
        },
        mimetype='text/event-stream',
    )


@bp.route("/notifications")#, methods=["POST"])
@auth.login_required
def notifications():
    user = auth.current_user()
    toast = db.session.query(ToastNotifications).filter_by(
        seen_at=None).filter_by(user_id=user.id).filter_by(company_id=user.company_id).first()
    if toast:
        if toast.type == "badge":
            badge = db.session.query(GamificationBadge).filter_by(id=toast.message).first()
            if badge:
                messenger.announce_badge(badge=badge, duration=5000)
            else:
                logger.info(f"Badge id {str(toast.message)} not found.")
        else:
            msg = json.dumps({
                'text': toast.message,
                'autoHideDuration': 3000,
                'variant': toast.type,
                'anchorOrigin': {
                    'horizontal': 'left',
                    'vertical': 'bottom',
                }
            })
            messenger.announce(f'data: {msg}\n\n')
        toast.set_seen()
        return Response(response=f"New Notifications found", status=200)
    return Response(response=f"No Notification", status=200)



from . import bp
from polzybackend import messenger, db
from polzybackend.models import User, GamificationBadge, GamificationBadgeType, GamificationBadgeLevel
from flask import jsonify
import json
from random import sample, choice

# toast debugging route
@bp.route('/ping')
def ping():
    #msg = messenger.format_sse(data='ping', event='fasifu')
    data = {
        'text': 'Fasifu message!',
        #'variant': 'success',
        #'anchorOrigin': {
        #    'vertical': 'top',
        #    'horizontal': 'left',
        #},
        #'autoHideDuration': 2000,
    }
    msg = f"data: {json.dumps(data)}\n\n"
    messenger.announce(msg=msg)
    return {'success': msg}, 200


# new badge for admin in current organization
@bp.route('/newbadge')
def newbadge():

    # get user 
    user = User.query.first()
    if user is None:
        return {'error': 'There are no user in DB'}, 400

    # create random badge
    # get possible types
    badge_level = GamificationBadgeLevel.query.first()
    all_types = GamificationBadgeType.query.all()
    search_type = True
    count = 0
    while search_type:
        badge_type = choice(all_types)
        search_type = any(b.type == badge_type for b in user.badges)
        if search_type:
            current_badge = next(filter(lambda b: b.type == badge_type, user.badges))
            #search_type = any(b.type == badge_type and b.level.next_level is None for b in user.badges)
            if current_badge.level.next_level:
                badge_level = current_badge.level.next_level
                new_badge = current_badge
                new_badge.level = badge_level
                new_badge.is_seen = False
                search_type = False
        else:
            new_badge = GamificationBadge(
                user=user,
                company=user.company.company,
                type=badge_type,
                level=badge_level,
            )
            db.session.add(new_badge)

        # break if all user badges of the highest level
        if count > 100:
            raise Exception('No new badge is available:(')
    
    db.session.commit()

    messenger.announce_badge(new_badge)
    return {'success': 'OK'}, 200


# returns all available users
@bp.route('/users')
def get_users():
    users = User.query.all()
    return jsonify([u.email for u in users]), 200




from polzybackend.utils import generate_id
import uuid
from copy import deepcopy


class Policy:
    #
    # Policy details, activities, attributes 
    #

    def __init__(self, number, date):
        #
        # initialization
        #

        self.id = generate_id()
        self.number = number
        self.effective_date = date
        self.data = None
        self.user = None

    def set_user(self, user):
        self.user = deepcopy(user)

    def setStage(self, stage):
        self.stage = stage

    def setLanguage(self, language):
        self.language = language

    def fetch(self):
        #
        # IMPORTANT: this method should be define within custon implementation
        #
        # fetches policy details from Policy Management System
        #

        raise Exception('Fetch method is not set in Policy class')

    def executeActivity(self, data):
        #
        # IMPORTANT: this method should be define within custon implementation
        #
        # executes policy activity defined in data
        #

        raise Exception('Execute Activity method is not set in Policy class')

    def get(self):
        #
        # returns Policy data
        #

        result = {'id': self.id}
        if self.data:
            result.update(self.data)

        return result


class Antrag:
    #
    # PoLZy interface instance for antrag
    #

    def __init__(self, product_name, user):
        self.id = generate_id()
        self.product_name = product_name
        self.user = deepcopy(user)
        self.instance = None


    def clone(self):
        antrag_copy = deepcopy(self)
        antrag_copy.id = generate_id()

        return antrag_copy


    def initialize(self):
        #
        # IMPORTANT: this method should be define within custom implementation
        #
        # initializes antrag instance within Policy Management System 
        #

        raise Exception('Initialize method is not defined in Antrag class')


    def get(self):
        #
        # IMPORTANT: this method should be define within custom implementation
        #
        # returns antrag instance as json object to front-end 
        #

        raise Exception('Get method is not defined in Antrag class')
 

    def updateFields(self, data):
        #
        # IMPORTANT: this method should be define within custom implementation
        #
        # updates antrag fields based on data
        #

        raise Exception('Update Fields method is not defined in Antrag class')


    def executeActivity(self, data):
        #
        # IMPORTANT: this method should be define within custom implementation
        #
        # executes antrag activity defined in data
        #

        raise Exception('Execute Activity method is not defined in Antrag class')




from flask import jsonify, request, current_app, send_file, abort
from polzybackend.antrag import bp
from polzybackend.utils.import_utils import antrag_products, antrag_class
from polzybackend import auth

@bp.route('/antrag/products')
@auth.login_required
def get_antrag_products():
    #
    # returns all products that available for current user
    #
    try:
        # get all stages
        product_list = antrag_products().getAllAvailableProducts(auth.current_user())
        return jsonify(product_list), 200

    except Exception as e:
        current_app.logger.exception(f"Error during get_antrag_products: {e}")
        
    return jsonify({'error': f'Failed to get antrag products'}), 400


@bp.route('/antrag/new/<string:product_type>')
@auth.login_required
def new_antrag(product_type):
    #
    # creates an Antrag by product_type
    # and returns it to frontend
    #
    try:
        antrag = antrag_class()(product_type, auth.current_user())

        # store antrag and return it to store and return json object
        antrag.initialize()
        current_app.config['ANTRAGS'][antrag.id] = antrag
        result = antrag.get()
        return jsonify(result), 200

    except Exception as e:
        current_app.logger.exception(f'Initialization of antrag instance {product_type} failed: {e}')

    return jsonify({'error': f'Initialization of antrag instance failed'}), 400


@bp.route('/antrag/clone/<string:id>')
@auth.login_required
def clone_antrag(id):
    #
    # clone antrag instance by id
    #

    try:
        # get antrag from app store
        antrag_src = current_app.config['ANTRAGS'].get(id)
        if antrag_src is None:
            raise Exception(f'Antrag {id} not found')

        # make copy and return antrag json object
        antrag = antrag_src.clone()
        current_app.config['ANTRAGS'][antrag.id] = antrag
        result = antrag.get()
        return jsonify(result), 200

    except Exception as e:
        current_app.logger.warning(f'Cloning of antrag {id} failed: {e}')
    
    return jsonify({'error': f'Cloning of antrag instance failed'}), 400


@bp.route('/antrag/update', methods=['POST'])
@auth.login_required
def update_antrag():
    #
    # updates antrag fields
    #

    # get post data
    data = request.get_json()

    # get antrag and update its values
    try:
        # get antrag from app store
        antrag = current_app.config['ANTRAGS'].get(data['id'])
        if antrag is None:
            raise Exception(f'Antrag {data["id"]} is absent in PoLZy storage. Most probably instance restarted.')

        # update antrag values and return antrag json object
        antrag.updateFields(data)
        result = antrag.get()
        return jsonify(result), 200

    except Exception as e:
        current_app.logger.exception(f'Antrag {data["id"]}, fields update failed: {e}')
    
    return jsonify({'error': f'Update of the antrag fields failed'}), 400


@bp.route('/antrag/execute', methods=['POST'])
@auth.login_required
def execute_antrag():
    #
    # executes antrag activities
    #

    # get post data
    data = request.get_json()

    # get antrag and execute activity
    try:
        # get antrag from app store
        antrag = current_app.config['ANTRAGS'].get(data['id'])
        if antrag is None:
            raise Exception(f'Antrag {data["id"]} is absent in PoLZy storage. Most probably instance restarted.')

        # execute antrag activity and return the response object
        result = antrag.executeActivity(data)
        return jsonify(result), 200

    except Exception as e:
        current_app.logger.exception(f'Failed execute activity {data.get("activity")} of antrag {data["id"]}: {e}')
    
    return jsonify({'error': f'Execution of antrag activiy {data.get("activity")} failed'}), 400



from polzybackend import db, auth
from polzybackend.utils import generate_id, date_format
from polzybackend.utils.auth_utils import generate_token, get_expired, is_supervisor, load_attributes
from datetime import datetime, date
from sqlalchemy import and_, or_
from functools import reduce
import json


# authentication
@auth.verify_token
def verify_token(token):
    user = User.query.filter_by(access_key=token).first()
    if user and user.key_expired > datetime.utcnow():
        return user


#
# Authentication Models
#

#
# Association Models
#

# association table for user-role relation
user_company_roles = db.Table(
    'user_company_roles',
    db.Column('user_id', db.String(56), db.ForeignKey('users.id'), primary_key=True),
    db.Column('company_id', db.String(56), db.ForeignKey('companies.id'), primary_key=True),
    db.Column('role_id', db.Integer, db.ForeignKey('roles.id'), primary_key=True)
)


class CompanyToCompany(db.Model):
    #
    # association object between companies
    #
    __tablename__ = 'company_company'
    parent_id = db.Column(db.String(56), db.ForeignKey('companies.id'), primary_key=True)
    child_id = db.Column(db.String(56), db.ForeignKey('companies.id'), primary_key=True)
    attributes = db.Column(db.String(1024), nullable=True)

    # relationships
    parent = db.relationship('Company', backref='child_companies', foreign_keys=[parent_id])
    child = db.relationship('Company', backref='parent_companies', foreign_keys=[child_id])


class UserToCompany(db.Model):
    #
    # association object between user and company
    #
    __tablename__ = 'user_company'
    user_id = db.Column(db.String(56), db.ForeignKey('users.id'), primary_key=True)
    company_id = db.Column(db.String(56), db.ForeignKey('companies.id'), primary_key=True)
    attributes = db.Column(db.String(1024), nullable=True)

    # relationships
    user = db.relationship(
        'User',
        lazy='subquery',
        backref='companies',
        foreign_keys=[user_id],
    )
    company = db.relationship(
        'Company',
        lazy='subquery',
        backref='users',
        foreign_keys=[company_id],
    )
    roles = db.relationship(
        'Role',
        lazy='subquery',
        secondary=user_company_roles,
        primaryjoin=and_(user_id == user_company_roles.c.user_id, company_id == user_company_roles.c.company_id),
        backref=db.backref('user_to_companies'),
    )
    #badges = db.relationship(
    #    'GamificationBadge',
    #    lazy='subquery',
    #    primaryjoin="and_(UserToCompany.user_id==GamificationBadge.user_id, "
    #                     "UserToCompany.company_id==GamificationBadge.company_id)"
    #)


    def to_json(self):
        return {
            'id': self.company.id,
            'name': self.company.name,
            'displayedName': str(self.company),
            'attributes': load_attributes(self.company.attributes),
        }

    def to_admin_json(self):
        child_companies = CompanyToCompany.query.filter_by(parent_id=self.company_id).all()
        return {
            'id': self.company.id,
            'name': self.company.name,
            'displayedName': str(self.company),
            'childCompanies': [{
                'id': ch.child_id,
                'name': ch.child.name,
                'attributes': ch.attributes,
            } for ch in child_companies],
            'users': [{
                'id': u.user.id,
                'email': u.user.email,
                'roles': [r.name for r in u.roles],
                'attributes': u.attributes if u.attributes else '',
            } for u in self.company.users],
        }


class Role(db.Model):
    __tablename__ = 'roles'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(16), unique=True, nullable=False)
    is_supervisor = db.Column(db.Boolean, nullable=False, default=False)


class User(db.Model):
    __tablename__ = 'users'
    id = db.Column(db.String(56), primary_key=True, default=generate_id)
    email = db.Column(db.String(64), unique=True, nullable=False)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    displayed_name = db.Column(db.String(64), nullable=True)
    first_name = db.Column(db.String(64), nullable=True)
    last_name = db.Column(db.String(64), nullable=True)
    # oauth_provider_id = db.Column(db.Integer, db.ForeignKey('oauth_providers.id'), nullable=False)
    # oauth_user_id = db.Column(db.String(128), nullable=False)
    # oauth_token = db.Column(db.String(128), nullable=False)
    access_key = db.Column(db.String(128), nullable=False, default=generate_token)
    key_expired = db.Column(db.DateTime, nullable=False, default=get_expired)
    # current session attributes
    stage = db.Column(db.String(8), nullable=True)
    language = db.Column(db.String(8), nullable=True)
    company_id = db.Column(db.String(56), db.ForeignKey('companies.id'), nullable=True)

    # relationships
    # oauth_provider = db.relationship(
    #    'OAuthProvider',
    #    foreign_keys=[oauth_provider_id],
    # )
    company = db.relationship(
        'UserToCompany',
        primaryjoin=and_(
            id == UserToCompany.user_id,
            company_id == UserToCompany.company_id,
        ),
        uselist=False,
        lazy='subquery',
    )
    badges = db.relationship(
        'GamificationBadge',
        primaryjoin="and_(User.id==GamificationBadge.user_id, "
                         "User.company_id==GamificationBadge.company_id)"
    )

    def __str__(self):
        return self.displayed_name or self.first_name or self.last_name or self.email.split('@')[0]

    def set_stage(self, stage):
        self.stage = stage
        db.session.commit()

    def set_language(self, language):
        self.language = language
        db.session.commit()

    def set_company(self, company_id=None, company=None):
        # check if company data is provided
        if company_id is None and company is None:
            raise Exception('Neither Company ID nor Company provided')

        # check if company should be fetched
        if company is None:
            company = Company.query.get(company_id)
        
        # check if company exists
        if company is None:
            raise Exception('Company not found')

        # get UserToCompany instance
        user_company = UserToCompany.query.filter(and_(
            UserToCompany.user_id == self.id,
            UserToCompany.company_id == company.id,
        )).first()

        # check if association exists
        if user_company is None:
            raise Exception(f'User is not assigned to company ID {company_id}')

        # set company
        self.company_id = company.id
        db.session.commit()

        # return company json
        return user_company.to_json()

    def is_supervisor(self):
        return reduce(lambda result, company: result or is_supervisor(company), self.companies, False)

    def get_permissions(self):
        if self.company:
            user_roles = [r.name.lower() for r in self.company.roles]

            return {
                'policy': any(role in ['admin', 'clerk'] for role in user_roles),
                'antrag': any(role in ['admin', 'agent'] for role in user_roles),
            }

        return {
            'policy': False,
            'antrag': False,
        }

    def to_json(self):
        return {
            'id': self.id,
            'email': self.email,
            'name': str(self),
            'isSupervisor': self.is_supervisor(),
            'stage': self.stage,
            'accessToken': self.access_key,
            'companies': [company.to_json() for company in self.companies],
        }

    def get_admin_json(self):
        return {
            'possibleRoles': [r.name for r in Role.query.all()],
            'companies': [company.to_admin_json() for company in filter(
                lambda company: is_supervisor(company),
                self.companies
            )],
        }


'''
class OAuthProvider(db.Model):
    __tablename__ = 'oauth_providers'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(64), unique=True, nullable=False)
    client_id = db.Column(db.String(128), nullable=False)
    secret_key = db.Column(db.String(128), nullable=False)

    def __str__(self):
        return self.name
'''


#
# Company Model
#
class Company(db.Model):
    __tablename__ = "companies"
    id = db.Column(db.String(56), primary_key=True, default=generate_id)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    name = db.Column(db.String(128), unique=True, nullable=False)
    displayed_name = db.Column(db.String(64), nullable=True)
    email = db.Column(db.String(32), nullable=True)
    phone = db.Column(db.String(16), nullable=True)
    country = db.Column(db.String(32), nullable=True)
    post_code = db.Column(db.String(8), nullable=True)
    city = db.Column(db.String(32), nullable=True)
    address = db.Column(db.String(64), nullable=True)
    attributes = db.Column(db.String(1024), nullable=True)

    def __str__(self):
        return self.displayed_name or self.name


#
# Activity Models
#
class Activity(db.Model):
    __tablename__ = 'activities'
    id = db.Column(db.String(56), primary_key=True, default=generate_id)
    creator_id = db.Column(db.String(56), db.ForeignKey('users.id'), nullable=False)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    policy_number = db.Column(db.String(64), nullable=False)
    effective_date = db.Column(db.Date, nullable=False, default=date.today)
    type = db.Column(db.String(32), nullable=False)
    status = db.Column(db.String(64), nullable=True)
    finished = db.Column(db.DateTime, nullable=True)
    attributes = db.Column(db.String, nullable=True)
    
    # relationships
    creator = db.relationship(
        'User',
        backref=db.backref('created_activities', order_by='desc(Activity.created)'),
        foreign_keys=[creator_id],
    )

    def __str__(self):
        return self.id

    @classmethod
    def new(cls, data, policy, current_user):
        # 
        # create instance using data
        #

        instance = cls(
            policy_number=policy.number,
            effective_date=datetime.strptime(policy.effective_date, date_format).date(),
            type=data['activity'].get('name'),
            creator_id=current_user.id,
            attributes=json.dumps(data['activity'].get('fields'))
        )
        
        # store to db
        db.session.add(instance)
        db.session.commit()
        
        return instance

    @classmethod
    def read_policy(cls, policy_number, effective_date, current_user):
        #
        # create instance of reading a policy
        #

        instance = cls(
            policy_number=policy_number,
            effective_date=datetime.strptime(effective_date, date_format).date(),
            type='Read Policy',
            creator=current_user,
            finished=datetime.utcnow(),
            status='OK',
        )

        # store to db
        db.session.add(instance)
        db.session.commit()
        
        return instance

    def finish(self, status):
        #
        # sets is_finished to True 
        #
        self.status = status
        self.finished = datetime.utcnow()
        db.session.commit()


#
# Gamification Models
#
class GamificationEvent(db.Model):
    __tablename__ = 'gamification_events'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(32), unique=True, nullable=False)

    def __str__(self):
        return self.name


class GamificationActivity(db.Model):
    __tablename__ = 'gamification_activities'
    id = db.Column(db.Integer, primary_key=True)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    processed = db.Column(db.DateTime, nullable=True)
    is_processed = db.Column(db.Boolean, nullable=False, default=False)
    event_details = db.Column(db.String(1024), nullable=True)
    event_id = db.Column(db.Integer, db.ForeignKey('gamification_events.id'), nullable=False)
    user_id = db.Column(db.String(56), db.ForeignKey('users.id'), nullable=False)
    company_id = db.Column(db.String(56), db.ForeignKey('companies.id'), nullable=False)

    # relationships
    event = db.relationship('GamificationEvent', foreign_keys=[event_id])
    user = db.relationship('User', backref='gamification_activities', foreign_keys=[user_id])
    company = db.relationship('Company', backref='gamification_activities', foreign_keys=[company_id])

    def __str__(self):
        msg = f'{self.user.email}: {self.event.name}'
        if self.processed:
            return f'{msg} - processed at {self.processed.strftime("%d.%m.%Y, %H:%M:%S")}'
        else:
            return f'{msg} - NOT processed'

    @classmethod
    def new(cls, user, event, event_details):
        # 
        # create instance based on the provided data
        #

        if not isinstance(event, GamificationEvent):
            if isinstance(event, int):
                event = db.session.query(GamificationEvent).filter_by(id=event).first()
            elif isinstance(event, str):
                event = db.session.query(GamificationEvent).filter_by(name=event).first()

        instance = cls(
            user_id=user.id,
            company_id=user.company.company_id,
            event=event,
            event_details=event_details,
        )

        db.session.add(instance)
        db.session.commit()
        
        return instance

    def set_processed(self, commit=True):
        #
        # set activity as processed
        #

        self.processed = datetime.utcnow()
        self.is_processed = True

        # save to db
        if commit:
            db.session.commit()


class GamificationUserStats(db.Model):
    __tablename__ = 'gamification_statistics'
    user_id = db.Column(db.String(56), db.ForeignKey('users.id'), primary_key=True, nullable=False)
    company_id = db.Column(db.String(56), db.ForeignKey('companies.id'), primary_key=True, nullable=False)
    type_id = db.Column(db.Integer, db.ForeignKey('gamification_badge_types.id'), primary_key=True)
    daily = db.Column(db.Integer, default=0)
    weekly = db.Column(db.Integer, default=0)
    monthly = db.Column(db.Integer, default=0)
    yearly = db.Column(db.Integer, default=0)
    all_time = db.Column(db.Integer, default=0)
    last_updated = db.Column(db.DateTime, default=datetime.now, onupdate=datetime.now)

    user = db.relationship('User', backref='gamification_statistics', foreign_keys=[user_id])
    company = db.relationship('Company', backref='gamification_statistics', foreign_keys=[company_id])
    type = db.relationship('GamificationBadgeType', backref="gamification_badge_types", foreign_keys=[type_id])

    def __str__(self):
        return self.user_id + f" - {self.type_id} -" + " -- " + str(self.get_user_statistics())

    def check_timeline(self):
        year_is_changed = self.last_updated.date().year < datetime.now().date().year
        # year_is_changed used when year changed but month or week number is same
        changes = False
        if self.last_updated.date() < datetime.now().date() or year_is_changed:
            self.daily = 0
            changes = True
        if self.last_updated.date().isocalendar()[1] < datetime.now().date().isocalendar()[1] or year_is_changed:
            self.weekly = 0
            changes = True
        if self.last_updated.date().month < datetime.now().date().month or year_is_changed:
            self.monthly = 0
            changes = True
        if year_is_changed:
            self.yearly = 0
            changes = True
        if changes:  # only update last_update time when it is needed
            self.last_updated = datetime.now()
            db.session.commit()

    def add_points(self, points=1, weight=1):
        self.check_timeline()
        weighted_point = points * weight
        self.daily += weighted_point
        self.weekly += weighted_point
        self.monthly += weighted_point
        self.yearly += weighted_point
        self.all_time += weighted_point
        self.update_badge()
        return {"all time": self.all_time}

    def get_user_statistics(self):
        self.check_timeline()
        json_data = {"daily": self.daily, "weekly": self.weekly, "monthly": self.monthly,
                     "annual": self.yearly, "all time": self.all_time}
        return json_data

    @classmethod
    def new(cls, user_id, company_id, type_id):
        instance = cls(
            user_id=user_id,
            company_id=company_id,
            type_id=type_id,
        )
        db.session.add(instance)
        db.session.commit()

        return instance

    def update_badge(self):
        level_id = self.get_level_id()
        badge = db.session.query(GamificationBadge).filter_by(user_id=self.user_id).filter_by(
            company_id=self.company_id).filter_by(type_id=self.type_id).filter_by(level_id=level_id).first()
        if not badge:
            badge = GamificationBadge(
                user_id=self.user_id, company_id=self.company_id, type_id=self.type_id, level_id=level_id)
            db.session.add(badge)
            db.session.commit()
            ToastNotifications.new(
                message=badge.id, type="badge", company_ids=self.company_id, user_ids=self.user_id
            )

    def get_level_id(self):
        level = db.session.query(GamificationBadgeLevel).filter(and_(self.all_time >= GamificationBadgeLevel.min_level,
            (or_(self.all_time <= GamificationBadgeLevel.max_level, GamificationBadgeLevel.max_level == None)))).first()
        return level.id

    @staticmethod
    def get_type_id(event: GamificationEvent, event_details):
        eventName = event.name.lower()
        event_details = json.loads(event_details)
        eventType = event_details.get("lineOfBusiness")
        if "policy" in eventName and eventType:
            return db.session.query(GamificationBadgeType).filter_by(name=f"Polizze {eventType}").first().id
        elif "antrag" in eventName and eventType:
            return db.session.query(GamificationBadgeType).filter_by(name=f"Antrag {eventType}").first().id
        elif "login" in eventName:
            return db.session.query(GamificationBadgeType).filter_by(name="Login").first().id
        elif "policy" in eventName:
            return db.session.query(GamificationBadgeType).filter_by(name=f"Policy").first().id
        elif "antrag" in eventName:
            return db.session.query(GamificationBadgeType).filter_by(name=f"Antrag").first().id
        else:
            print(f"{eventName} event is ignored.")
            return None

    @staticmethod
    def get_weight(event: GamificationEvent, event_details):
        eventName = event.name
        event_details = json.loads(event_details)
        lob = event_details.get("lineOfBusiness", "")
        activityName = event_details.get("Activity")
        if lob and activityName:
            id_ = db.session.query(GamificationActivityWeight).filter(and_(
                GamificationActivityWeight.activity_name == activityName,
                GamificationActivityWeight.line_of_business == lob)).first()
        else:
            if lob and eventName:
                id_ = db.session.query(GamificationActivityWeight).filter(and_(
                    GamificationActivityWeight.activity_name == eventName,
                    GamificationActivityWeight.line_of_business == lob)).first()
            else:
                id_ = db.session.query(GamificationActivityWeight).filter_by(activity_name=eventName).first()
        if not id_:
            print(f"Combination of activityName: {str(activityName)}, lineOfBusiness: {str(lob)} or "
                  f"event: {str(eventName)} not found in Weight table. Using default value 10.")
            id_ = db.session.query(GamificationActivityWeight).filter_by(activity_name="default").first()
            if not id_:
                print("Default weight not found in database. Returning 10")
                return 10
        return id_.points

    @classmethod
    def create_or_update_row(cls, activity: GamificationActivity, commit=True):
        type_id = cls.get_type_id(activity.event, activity.event_details)
        weightage = cls.get_weight(activity.event, activity.event_details)
        if not type_id:
            return None
        user_id = activity.user_id
        company_id = activity.company_id
        user_stats = db.session.query(GamificationUserStats).filter_by(user_id=user_id
            ).filter_by(company_id=company_id).filter_by(type_id=type_id).first()
        if user_stats:
            user_stats.add_points(weight=weightage)
        if not user_stats:
            user_stats = GamificationUserStats.new(user_id=user_id, company_id=company_id, type_id=type_id)
            user_stats.add_points(weight=weightage)
        if commit:
            db.session.commit()
        return user_id

    @staticmethod
    def commit():
        db.session.commit()

## Badges

class GamificationBadgeDescription(db.Model):
    #
    # Badge earn requirements for given Type and Level
    #
    __tablename__ = 'gamification_badge_descriptions'
    type_id = db.Column(db.Integer, db.ForeignKey('gamification_badge_types.id'), primary_key=True)
    level_id = db.Column(db.Integer, db.ForeignKey('gamification_badge_levels.id'), primary_key=True)
    description = db.Column(db.String(512), nullable=False)

    # relationships
    type = db.relationship('GamificationBadgeType', foreign_keys=[type_id])
    level = db.relationship('GamificationBadgeLevel', foreign_keys=[level_id])

class GamificationBadgeLevel(db.Model):
    __tablename__ = 'gamification_badge_levels'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(32), unique=True, nullable=False)
    min_level = db.Column(db.Integer, unique=True, nullable=False)
    max_level = db.Column(db.Integer, unique=True, nullable=True)
    is_lowest = db.Column(db.Boolean, nullable=False, default=False,)
    next_level_id = db.Column(db.Integer, db.ForeignKey('gamification_badge_levels.id'), nullable=True)

    #relationships
    next_level = db.relationship('GamificationBadgeLevel', remote_side=[next_level_id], uselist=False)

    def __str__(self):
        return self.name

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.max_level = self.get_next_level_min()

    def get_next_level_name(self):
        if self.next_level:
            return self.next_level.name

    def get_next_level_min(self):
        if self.next_level:
            return self.next_level.min_level - 1
        return None


class GamificationBadgeType(db.Model):
    __tablename__ = 'gamification_badge_types'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(32), unique=True, nullable=False)
    title = db.Column(db.String(32), nullable=False)

    # reletionaships
    descriptions = db.relationship(
        'GamificationBadgeDescription',
        primaryjoin="GamificationBadgeType.id==GamificationBadgeDescription.type_id",
    )

    def get_description(self):
        return {(d.level.name if not d.level.is_lowest else 'lowest'): d.description for d in self.descriptions}

    def __str__(self):
        return self.name

    def to_json(self):
        return {
            'name': self.name,
            'title': self.title,
            'description': self.get_description(),
        }

class GamificationBadge(db.Model):
    __tablename__ = 'gamification_badges'
    id = db.Column(db.Integer, primary_key=True)  
    user_id = db.Column(db.String(56), db.ForeignKey('users.id'), nullable=False)
    company_id = db.Column(db.String(56), db.ForeignKey('companies.id'), nullable=False)
    type_id = db.Column(db.Integer, db.ForeignKey('gamification_badge_types.id'), nullable=False)
    level_id = db.Column(db.Integer, db.ForeignKey('gamification_badge_levels.id'), nullable=False)
    achieved_at = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    is_seen = db.Column(db.Boolean, nullable=False, default=False)

    # relationships
    user = db.relationship('User', backref='gamification_badges', foreign_keys=[user_id])
    company = db.relationship('Company', foreign_keys=[company_id])
    type = db.relationship('GamificationBadgeType', foreign_keys=[type_id])
    level = db.relationship('GamificationBadgeLevel', foreign_keys=[level_id])

    def __str__(self):
        return f'Badge {self.level.name} {self.type.name} for {self.user.email}'

    def set_seen(self):
        self.is_seen = True
        db.session.commit()

    def to_json(self):
        return {
            'type': self.type.name,
            'level': self.level.name,
            'next_level': self.level.get_next_level_name(),
            'isSeen': self.is_seen,
        }

    @classmethod
    def get_user_stats(cls, user_id=user_id):
        user = db.session.query(cls).filter_by(user_id=user_id).all()
        json_data = [data.to_json() for data in user]
        return json_data


class GamificationActivityWeight(db.Model):
    __table_name__ = "gamification_activity_weight"
    __table_args__ = (db.UniqueConstraint('activity_name', 'line_of_business', name='unique'),)
    id = db.Column(db.Integer, primary_key=True)
    activity_name = db.Column(db.String(32), nullable=False)
    line_of_business = db.Column(db.String(16), nullable=True)
    points = db.Column(db.Integer, nullable=False)

    @classmethod
    def get_points(cls, id):
        if id:
            weight = cls.query.filter(cls.id == id).first()
            if weight:
                return weight.points
        print("Weightage id is either not supplied or not found in Weightage table. Returned default value of 1")
        return 1

    @classmethod
    def new(cls, activity_name, line_of_business, points=1):
        instance = cls(activity_name=activity_name, line_of_business=line_of_business, points=points)
        db.session.add(instance)
        db.session.commit()

    def to_json(self):
        js = {
            "id": self.id,
            "activity_name": self.activity_name,
            "line_of_business": self.line_of_business,
            "points": self.points
        }
        return js

    @classmethod
    def get_all_id_points(cls):
        all_rows = cls.query.all()
        all_data = {row.id: row.points for row in all_rows}
        return all_data


class ToastNotifications(db.Model):
    __tablename__ = "toast_notifications"
    id = db.Column(db.String(56), primary_key=True, default=generate_id)
    company_id = db.Column(db.String(56), db.ForeignKey('companies.id'), primary_key=True)
    user_id = db.Column(db.String(56), db.ForeignKey('users.id'), primary_key=True)
    message = db.Column(db.String(128), nullable=False)
    type = db.Column(db.String(16), default="default", nullable=False)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    seen_at = db.Column(db.DateTime)

    @classmethod
    def new(cls, message, type="default", company_ids=None, user_ids=None):
        companies = []
        if company_ids:
            if isinstance(company_ids, list):
                for company_id in company_ids:
                    companies.append(company_id)
            else:
                companies.append(company_ids)
        if not companies:
            companies = [company.id for company in db.session.query(Company).all()]
        for company_id in companies:
            users = []
            if not user_ids:
                users = [user.id for user in db.session.query(User).filter_by(company_id=company_id).all()]
            else:
                if isinstance(user_ids, list):
                    for user_id in user_ids:
                        users.append(user_id)
                else:
                    users.append(user_ids)
            for user_id in users:
                instance = cls(company_id=company_id, user_id=user_id, message=message, type=type)
                db.session.add(instance)
        db.session.commit()

    def set_seen(self):
        self.seen_at = datetime.now()
        db.session.commit()



import queue
import json

class Messenger:

    def __init__(self):
        self.listeners = []

    def listen(self):
        q = queue.Queue(maxsize=5)
        self.listeners.append(q)
        return q

    def announce(self, msg):
        for i in reversed(range(len(self.listeners))):
            try:
                self.listeners[i].put_nowait(msg)
            except queue.Full:
                del self.listeners[i]

    def announce_badge(self, badge, text=None, anchors=None, duration=None):
        # set toast props
        msg_data = {
            'uri': f'/{badge.type.name.lower()}/{badge.level.name.lower()}',
            'text': text or f'{badge.level.name} Badge {badge.type.title} Achieved',
            'anchorOrigin': anchors or {
                'vertical': 'top',
                'horizontal': 'left',
            },
            'autoHideDuration': duration,
        }

        # get the message
        message = self.format_dict(
            event='newbadge',
            data=msg_data,
        )
        
        self.announce(message)

    @classmethod
    def format_sse(cls, data: str, event=None) -> str:
        msg = f'data: {data}\n\n'
        if event is not None:
            msg = f'event: {event}\n{msg}'
        return msg

    @classmethod
    def format_dict(cls, data: dict, event=None) -> str:
        msg = f'data: {json.dumps(data)}\n\n'
        if event is not None:
            msg = f'event: {event}\n{msg}'
        return msg



from random import choice
from datetime import datetime, timedelta
from functools import reduce
import uuid
import string
import json


ACCESS_KEY_LENGTH = 64
EXPIRED_HOURS = 24


def generate_token(length=ACCESS_KEY_LENGTH):
    simbols = string.ascii_letters + string.digits + '!#$%&*+-<=>?@'
    return ''.join(choice(simbols) for i in range(length))


def get_expired(hours=EXPIRED_HOURS, days=0):
    return datetime.utcnow() + timedelta(hours=hours, days=days)


def is_supervisor(company):
	return reduce(lambda result, role: result or role.is_supervisor, company.roles, False)

def load_attributes(attributes):
	try:
		return json.loads(attributes)
	except Exception:
		return None



import setuptools
import os

if __name__ == '__main__':
    # description
    with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "readme.md"), "r") as fh:
        long_description = fh.read()

    # requirements
    with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "requirements.txt"), "r") as f:
        requirements = f.read().splitlines()

    setuptools.setup(
        name="polzybackend",
        version="0.6.0",
        author="Bernhard Buhl",
        author_email="buhl@buhl-consulting.com.cy",
        description="Open source Insurance Policy Life Cycle Management",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://baangt.org",
        packages=setuptools.find_packages(),
        data_files=[],
        package_data={},
        install_requires=requirements,
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        include_package_data=True,
        python_requires='>=3.6',
    )



from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from flask_migrate import Migrate
from flask_httpauth import HTTPTokenAuth
import os
from polzybackend.messenger import Messenger
import flask_monitoringdashboard as dashboard


# initialization
db = SQLAlchemy()
migrate = Migrate()
auth = HTTPTokenAuth(scheme='Bearer')
messenger = Messenger()


def create_app(config=None):
    # create application
    app = Flask(__name__)
    # set default config
    app.config['SECRET_KEY'] = os.getenv('SECRET_KEY', default='secret!key')
    app.config['JSON_SORT_KEYS'] = False
    app.config['SQLALCHEMY_DATABASE_URI'] = os.getenv(
        'DATABASE_URL',
        default='sqlite:///' + os.path.join(os.path.abspath(os.path.dirname(__file__)), 'polzy.db'),
    )
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
    app.config.from_object(config)

    # policy store
    app.config['POLICIES'] = {}
    app.config['ANTRAGS'] = {}

    db.init_app(app)
    migrate.init_app(app, db)

    from polzybackend.authenticate import bp as bp_auth
    app.register_blueprint(bp_auth)

    from polzybackend.policy import bp as bp_policy
    app.register_blueprint(bp_policy)

    from polzybackend.antrag import bp as bp_antrag
    app.register_blueprint(bp_antrag)

    from polzybackend.administration import bp as bp_admin
    app.register_blueprint(bp_admin)

    from polzybackend.gamification import bp as bp_gamification
    app.register_blueprint(bp_gamification)

    from polzybackend.announce import bp as bp_announce
    app.register_blueprint(bp_announce)

    # -----------> DEBUG BLUEPRINT
    from polzybackend.debug import bp as bpDebug
    app.register_blueprint(bpDebug)

    # setup monitoring dashboard
    if app.config.get('DASHBOARD_CONFIG'):
        dashboard.config.init_from(file=app.config['DASHBOARD_CONFIG'])
    if app.config.get('DASHBOARD_DATABASE_URI'):
        dashboard.config.database_name = app.config['DASHBOARD_DATABASE_URI']
    dashboard.bind(app)

    return app


from polzybackend import models



from flask import jsonify, request, current_app
from datetime import date
from polzybackend.administration import bp
from polzybackend.models import User
from polzybackend import auth, db
from polzybackend.models import User, Company, Role, UserToCompany, CompanyToCompany
#from polzybackend.utils.auth_utils import str_id_to_bytes
from sqlalchemy import and_


@bp.route('/admin')
@auth.login_required
def get_admin_data():
    # returns admin data for companies where current user is admin
    try:
        result = auth.current_user().get_admin_json()
        return jsonify(result), 200
    except Exception as e:
        current_app.logger.warning(f'Failed to get admin data: {e}')
        return {'error': 'Failed to fetch data'}, 400


@bp.route('/admin/user-company/<string:action>', methods=['POST'])
@auth.login_required
def manage_user(action):
    #
    # manage user in company
    # <action> should be one of:
    # - add -> add user to company
    # - edit -> edit user roles in company
    # - remove -> remove user from company
    #

    try:
        # validate action
        if not action in ['add', 'edit', 'remove']:
            raise Exception(f"Action '{action}' is not supported") 
        
        # get request data
        data = request.get_json()
        
        # check for required fields in data
        if data.get('companyId') is None:
            raise Exception('Company ID is not defined in request')
        if data.get('email') is None:
            raise Exception('User e-mail is not defined in request')
        # roles are required by 'add' and 'edit' actions
        if (action in ['add', 'edit']) and (data.get('roles') is None):
            raise Exception(f'User roles are not defined in request. Required for action: {action}')

        # get company from db
        company = Company.query.get(data['companyId'])
        if company is None:
            raise Exception(f'Company with ID {data["companyId"]} not found')

        # get user from db
        user = User.query.filter_by(email=data['email']).first()
        # check if user exists
        if user is None:
            # create user for action 'add'
            if action == 'add':
                user = User(email=data['email'])
                db.session.add(user)
            # raise UserNotFound
            else:
                raise Exception(f'User {data["email"]} not found')

        # get or create user to company association
        if action in ['edit', 'remove']:
            # get
            user_to_company = UserToCompany.query.filter(and_(
                UserToCompany.company == company,
                UserToCompany.user == user,
            )).first()
            if user_to_company is None:
                raise Exception(f'User {data["email"]} not found in company {data["companyId"]}')
        else:
            # create
            user_to_company = UserToCompany(
                user=user,
                company=company,
            )
            db.session.add(user_to_company)

        # set user roles and attributes
        if action in ['add', 'edit']:
            # roles
            roles = []
            for name, value in data['roles'].items():
                if value:
                    # get role
                    role = Role.query.filter_by(name=name).first()
                    if role is None:
                        raise Exception(f"Role '{name}' not found")
                    roles.append(role)                
            # update user roles
            user_to_company.roles = roles

            # attributes
            if not data.get('attributes') is None:
                user_to_company.attributes = data['attributes']
        
        # remove user from company
        else:
            db.session.delete(user_to_company)

        # commit changes
        db.session.commit()
        
        # return updated admin data
        return jsonify(auth.current_user().get_admin_json()), 200
    
    except Exception as e:
        db.session.rollback()
        current_app.logger.warning(f"Failed to execute action '{action}'\nPayload: {request.get_json()}\nException: {e}")
        return {'error': str(e)}, 400


@bp.route('/admin/child-company/<string:action>', methods=['POST'])
@auth.login_required
def manage_child_company(action):
    #
    # manage child companies
    # <action> should be one of:
    # - add -> add user to company
    # - edit -> edit user roles in company
    # - remove -> remove user from company
    #

    try:
        # validate action
        if not action in ['add', 'edit', 'remove']:
            raise Exception(f"Action '{action}' is not supported") 
        
        # get request data
        data = request.get_json()
        
        # check for required fields in data
        if data.get('parentCompanyId') is None:
            raise Exception('Parent company ID is not defined in request')
        if data.get('childCompanyId') is None:
            raise Exception('Child company ID is not defined in request')
        if action == 'edit' and data.get('attributes') is None:
            raise Exception(f'Attributes are not defined in request. Required for action: {action}')

        # get companies from db
        parent_company = Company.query.get(data['parentCompanyId'])
        if parent_company is None:
            raise Exception(f'Company with ID {data["parentCompanyId"]} not found')

        child_company = Company.query.get(data['childCompanyId'])
        if child_company is None:
            raise Exception(f'Company with ID {data["childCompanyId"]} not found')

        # get company-to-company association
        company_company = CompanyToCompany.query.filter(and_(
            CompanyToCompany.parent == parent_company,
            CompanyToCompany.child == child_company,
        )).first()
        if company_company is None:
            raise Exception(f'Company ID {data["parentCompanyId"]} does not have child company ID {data["childCompanyId"]}')

        # update attributes
        if action == 'edit':
            company_company.attributes = data['attributes']

        if action == 'remove':
            db.session.delete(company_company)

        # commit changes
        db.session.commit()
        
        # return updated admin data
        return jsonify(auth.current_user().get_admin_json()), 200
    
    except Exception as e:
        db.session.rollback()
        current_app.logger.warning(f"Failed to execute action '{action}'\nPayload: {request.get_json()}\nException: {e}")
        return {'error': str(e)}, 400




from flask import jsonify, request, current_app
from datetime import date
from polzybackend.authenticate import bp
from polzybackend.utils.import_utils import all_stages, permissions
from polzybackend.models import User
from polzybackend import auth


@bp.route('/stages')
def stages():
    #
    # returns list of all available stages for login
    #

    try:
        # get all stages
        stages = all_stages()()
        current_app.logger.debug(f"Value of stages: {stages}")

    except Exception as e:
        current_app.logger.warning(f'Failed to get All Stages: {e}')
        stages = []

    return jsonify(stages), 200


@bp.route('/login', methods=['POST'])
def login():
    # get request data
    data = request.get_json()
    
    # check for required fields in data
    required_fields = [
        'email',
        'stage',
        'language',
    ]
    for field in required_fields:
        value = data.get(field)  
        if value is None:
            return {'error': f'User data does not contain {field}'}, 400

    # get user from db
    email = data['email']
    user = User.query.filter_by(email=email).first()
    # check if user found
    if user is None:
        return {'error': f'User {email} does not exist'}, 404

    # update current stage and language
    user.set_stage(data['stage'])
    user.set_language(data['language'])

    return jsonify(user.to_json()), 200


@bp.route('/permissions', methods=['POST'])
@auth.login_required
def user_permissions():
    # get request data
    try:
        data = request.get_json()

        # get company id
        company_id = data.get('id')
        if company_id is None:
            current_app.logger.warning(f'Payload does not contain company id. Payload: {data}')
            raise Exception('Payload does not contain company id')

        # update company
        user = auth.current_user()
        company_details = user.set_company(company_id=company_id)
        
        return jsonify({
            'permissions': permissions(user),
            'company': company_details,
        }), 200

    except Exception as e:
        return {'error': f'Failed to set company: {e}'}, 400
